// midi1.ino
// By Peter Churchyard @codewizard58
//
// 7/4/17
//

#include "play.h"
#include "softbits.h"
#include "midibits.h"
#include "soundbits.h"

#define MUXOUT 8
#define CLK 10
#define BIT4  11    // top bit of counter
#define MR 12
#define LED 13
//
#define OCT1 3    // +1 octave
#define OCT2 4    // +2 octave
#define OCT3 7    // -2 octave
//
#define NNOTES  6

// #define WDEBUG 1
int monophonic;

unsigned char prog0[] = {
  0, 1,   // power_on(1)
  2,      // input_a0
  8,      // output_a5
  1,      // power_off
  0xff
};

extern const unsigned char PROGMEM tune_1[];

extern void settempo(unsigned int n);
extern void adjusttempo(int n);

void setup() {
  // put your setup code here, to run once:
  pinMode(OCT1, OUTPUT);
  digitalWrite(OCT1, LOW);
  pinMode(OCT2, OUTPUT);
  digitalWrite(OCT2, LOW);
  pinMode(OCT3, OUTPUT);
  digitalWrite(OCT3, LOW);

  pinMode(CLK, OUTPUT);  // clk
  pinMode(BIT4, INPUT);

  pinMode(MUXOUT, INPUT);   // muxout

  pinMode(13, OUTPUT);
  digitalWrite(LED, HIGH);
  
  pinMode(MR, OUTPUT);
  digitalWrite(MR, LOW);
  
  soft_setup();
  midi_setup(0);  // 0 is rxtx, 1 is USB if available
//  mouse_setup();
#ifdef WDEBUG
  Serial.begin(9600);
#endif

  settempo(180);

  startTune( &tune_1[0], 0);
  repeatTune(1);  
}

/////////////////////////////////////////////////////////////////////////////////
#ifdef WDEBUG
void ws()
{
  Serial.print(" ");
}

void eol()
{
  Serial.println("");
}

void iprt(int n)
{
  Serial.print(n);
}

void lprt(long n)
{
  Serial.print(n);
}

void sprt(char *s)
{
  Serial.print(s);
}
#endif

////////////////////////////////////////////////////////////////////////////////
int ticks;
int beats;
int state;

int tmode;      // tempo mode
unsigned long mticks;
unsigned long mdelta;
unsigned long mlast;
int ltick;      // toggled by lfo value.
int ktick;      // set by key 12 press 

unsigned long tcount;     // tempo ticks
unsigned long tdelta = 500000L/24;   // 120 bpm

int curnote = 33;
int curchan = 0;
int octave;
int koctave;
int leds=0;
int ledsoff;

// arpeggiator vars
int arpmode=1;
int arpseq;
int arpdir;
int arpbeats;   // sync with beat or free run...
int arplength = 3;   
int shiftmode;
int shiftnote;
int arppat;   // pattern to play
//
int ktemp1;
int ktemp2;
//
int muxout;   // 1 if OUTPUT

/////////////////////////////////////////////////////////////

int outnote( )
{
  if( octave & 1){
    digitalWrite(OCT1, HIGH);
  }else{
    digitalWrite(OCT1, LOW);
  }
  if( octave & 2){
    digitalWrite(OCT2, HIGH);
  }else{
    digitalWrite(OCT2, LOW);
  }
  if( octave & 4){
    digitalWrite(OCT3, HIGH);
  }else{
    digitalWrite(OCT3, LOW);
  }

}


//////////////////////////////////////////////////////////////
#define LASTNOTEPRIORITY 1
#define LOWESTNOTEPRIORITY 0

int noteflags = LASTNOTEPRIORITY;

int notecnt;
char notedata[NNOTES+NNOTES];

int arpnotecnt;
char arpdata[NNOTES+NNOTES];
char arpflags[NNOTES+NNOTES];

//////////////////////////////////
int cmdmode;    // used to control what the keyboard state is..


// sets curnote and octave..
int noteOn_sub(int n, int vel, int chan)
{ 
  
  if( n < C0 || n > E6-1){
    return 0; 
  }
//  Serial.print("On ");Serial.println(n);

  if( n >= C6){
    curnote = 12+n-C6;
    octave = 3;
  }else if( n >= C0 && n < C2){
    curnote = (n % 12 );
    octave = 4;       // set -2
    if( n >= C1 ){
      octave |= 1;    // set +1
    }
  }else if( n == C3){
    octave = 0;
    curnote = 12;
  }else {
    n = n - C2;
    octave = n / 12;
    curnote = (n % 12 );
  }
  curchan = chan;
  curnote = curnote+curnote;
  outnote();

  return 0;
}

void prt_notes(char *msg)
{ int i;

#ifdef WDEBUG
sprt(msg);
iprt(notecnt);ws();iprt(curnote);ws();iprt(octave);sprt("[");
for(i=0; i < notecnt; i++){
    iprt(notedata[i]);sprt(", ");
}
sprt("]\n");
#endif

}

void moveupnotes(char *buf, int j, int cnt)
{ int i;
  char *vbuf = &buf[NNOTES];

  i = cnt -1;     // One before last
  if( i >= NNOTES-1){ // make sure it fits.
    i = NNOTES-2;
  }
  for(; i >= j; i--){
    buf[i+1] = buf[i];
    vbuf[i+1] = vbuf[i];
  }
}

void movedownnotes(char *buf, int j, int cnt)
{ int i;
  char *vbuf = &buf[NNOTES];

  for(i = j; i < cnt-1; i++)
  {
    buf[i] = buf[i+1];
    vbuf[i] = vbuf[i+1];
  }
  
}

int noteOn(int n, int vel, int chan)
{ int i,j;
  int nn = 0;

  if( n < 0 || n > E6){
    return 0;
  }

  for(i = 0; i < notecnt; i++){
    if( notedata[i] == n){
      break;
    }
  }
  if( i == notecnt){  // new note
    j = 0;
    nn = 1;
    if( noteflags == LASTNOTEPRIORITY){
      moveupnotes(notedata, j, notecnt);
      if( arpmode){
        moveupnotes(arpflags, j, notecnt);
        arpflags[0] = 0;
      }
    }else {
      // lowestnote priority
      for(i=0; i < notecnt; i++){
        if( n < notedata[i]){
          break;  // this is lower
        }
      }
      j = i;
      moveupnotes(notedata, j, notecnt);
      if( arpmode){
        moveupnotes(arpflags, j, notecnt);
        arpflags[0] = 0;
      }
      if( j > 0){
        nn = 0;   // not the lowest.. so no change
      }
    }
    // update the position
    notedata[j] = n;
    notedata[j+NNOTES] = vel;
    
    if( notecnt == 0 && shiftmode == 0){
      // start of new note sequence
      arpnotecnt = 0; // reset the arp note count
    }
    notecnt++;
    if( notecnt >= NNOTES){
      notecnt = NNOTES;
    }
  }
// if arp mode then let the doarp start playing..
  if( arpmode == 0){
    if( nn != 0 ){
      return noteOn_sub(notedata[0], notedata[NNOTES], chan);
    }
  }else {
    if( (arpmode & 1) || notecnt >= arpnotecnt){
      if( notecnt == 1){  // first note
        arpbeats = 0;
      }
      arpnotecnt = notecnt;

      for(i=0; i < NNOTES+NNOTES; i++){
        arpdata[i] = notedata[i];
      }
    }
    shiftnote = notedata[0] - arpdata[0];
  }
  return 0;
}



int noteOff_sub(int n, int chan)
{
  if( notecnt == 0){    // no notes playing
    curnote = 33;   // cannot match...
    return 0;
  }

//  prt_notes("Offx: ");
  if( arpmode == 0){
    noteOn_sub(notedata[0], notedata[0+NNOTES], 0);
  }
  
  return 0;
}

int noteOff(int n, int chan)
{ int i,j;

  for(i = 0; i < notecnt; i++){
    if( notedata[i] == n){
      break;
    }
  }
  if( i == notecnt){
//    sprt("Off not found ");iprt(n);eol();
    return 0;
  }

  movedownnotes(notedata, i, notecnt);

  notecnt--;
  if( notecnt < 0){
    notecnt = 0;
  }

  if( (arpmode & 1) == 1 ){
    arpnotecnt = notecnt;

    for(i=0; i < NNOTES+NNOTES; i++){
      arpdata[i] = notedata[i];
    }
  }

 return noteOff_sub(n, chan);

}

/////////////////////////////////////////////

#define NUMKEYS 20

int keystate[NUMKEYS];
int oldkeystate[NUMKEYS];

int cmdtimer = 0;
int curpatch;

void dopatch()
{
  
}

int koctavemap[16];


/////////////////////////////////////////////



int acnt;
int adata[6];
int tempo;
int ledmap[8] = { 1,8,2,4,0x80,0x20,0x40,0x10 };
int sendcc = 1;

int ctrlrange[] = {
  0,0,      // 0
  140, 750, // 1 lfo
  100,800,  // 2 freq
  220, 550, // 3 lfo wave
  100, 800, // 4 cutoff
  285, 600, // 5 pwm
};

int ctrlprev[6];
int ctrldata[6];

int ctrlchange(int ctrl, int val)
{   int n = ctrl + ctrl;
    int r = ctrlrange[n+1] - ctrlrange[n]+1;
    long midival = (val - ctrlrange[n]);

    if( midival < 0 ){
      midival = 0;
    }
    midival = (midival * 256) / r;

    if( midival > 255){
      midival = 255;
    }
    ctrldata[ ctrl] = midival;

    if( ctrl == 3){
      if( midival > 64){
        ltick = 1;
      }else {
        ltick = 0;
      }
    }
    if( tmode == 1 && ctrl == 5){
      adjusttempo(val);
    }
    
#ifdef WDEBUGX
  if( ctrl != 3){
Serial.print(ctrl);Serial.print(" ");Serial.println(val);
  }
#endif    
}

// scan the analog inputs.
//
int scanin()
{ int i;
  int v;
  int x;

  for(i=1; i < 6; i++){
    v = analogRead( i);
    x = adata[i] - v;
    if( x > 10 || x < -10){
        ctrlchange( i, v);
        adata[i] = v;
    }
  }

}


///////////////////////////////////////////////////////

void settempo(unsigned int n)
{
  tdelta = 60000000L / (n * 24);
//  iprt(n);ws();lprt(tdelta);eol();
}

int tadjust;

// 30 bpm = 83333
// 300 bpm = 8333

void adjusttempo(int n)
{
    n = n - 250+30;
    settempo(n);

    tmode = 1;
}

///////////////////////////////////////////////////////
// 

void doarp()
{ int x;
  int i;

  if( arpnotecnt == 0) // no note pressed.
  {
    arpbeats = 24;
    arpseq = 0;
    return;
  }
  arpbeats++;

  if( arpbeats == 1){
    if( (arpmode & 2) && shiftmode){
       noteOn_sub(arpdata[arpseq]+shiftnote, arpdata[arpseq+NNOTES], 0);
    }else {
       noteOn_sub(arpdata[arpseq], arpdata[arpseq+NNOTES], 0);
    }
  }

  if(arpbeats == arplength ){
    if( notecnt > 1){
      curnote = 37; // stop sounding
    }
  }
  if( arpbeats < 12){
    return;
  }

#ifdef WDEBUG
iprt(arpmode);ws();iprt(arpseq);ws(); iprt(arpdir);eol();
#endif
   if( (arpdir & 0x80) == 0x80){  // 'up'
    arpseq++;
    if( arpseq >= arpnotecnt){
      if( (arpmode & 0x04) && (arpnotecnt > 2)){
        arpdir &= 0x7f;
        arpseq -= 2;
      }else {
        arpseq = 0;
      }
    }
  }else {   // down
    arpseq--;
    if( arpseq < 0){
      if( (arpmode & 0x04) && (arpnotecnt > 2)){
        arpseq = 1;
        arpdir |= 0x80;
      }else {
        arpseq = arpnotecnt-1;
      }
    }
  }

  arpbeats = 0;

}

// 10 is clk
int count;
int cnt8;
int keyin;
int keymap[16];

int doclk()
{ int bit4;
  int keynote;
  int n;
  int next;
  

  // pulse the counter
  next = (count+1) & 0x1f;
  keynote = next/2;
  if( (next & 1)==0 || (keymap[ keynote] && (keynote+keynote) != curnote) ){
    if( muxout == 0){
      pinMode(MUXOUT, OUTPUT);
      muxout = 1;
    }
    digitalWrite(MUXOUT, LOW);  // do not sound
  }else {
    pinMode(MUXOUT, INPUT);
    muxout = 0;
  }
  
  digitalWrite(CLK, LOW);
  digitalWrite(CLK, HIGH);
  count = next;
  
  // read top bit of the counter
  bit4 = digitalRead( BIT4);
  // did it roll over back to 0?
  if( !bit4 && cnt8){
    if( count != 0){
      digitalWrite(LED, LOW);
    }
    count = 0;  // synchronise memory copy
  }else if( bit4 && !cnt8){
    count = 16; // update memory copy.
  }
  cnt8 = bit4;  // save for next time.
  
  if( count == 0){
    for(n=0; n < 16; n++){
      if( keymap[n]){
        break;
      }
    }
    if( n != 16){
      digitalWrite(LED, LOW);
    }else {
      digitalWrite(LED, HIGH);
    }
  }

  
  mticks = micros();
  if( mlast == 0){
    mlast = mticks;
  }
  mdelta = mticks - mlast;
  
  while( mdelta >= 10000){
    // lprt(mdelta);eol();
    if( mdelta > 100000L){
      mlast = mticks;
      mdelta = 10000;
    }
//    scanin();
    mlast = mlast + 10000;
    mdelta = mdelta - 10000;
    ticks++;

    tcount = tcount + 10000;
    while( tcount > tdelta){
//    iprt(tcount);eol();
      tcount -= tdelta;

      doarp();

      beats++;
      if( beats >= 24){ 
        beats = 0;

      }
    }
  }
  if( ticks > 10){
    ticks = 0;
  }

  keynote = count >> 1;
  if( muxout == 0){
    keyin = digitalRead(MUXOUT);
    if( keyin){
      
      if( keymap[keynote ] == 0){
        // first press
        noteOn(keynote+C2, 127, 0);
        keymap[keynote] = 1;
      }
        
    }else {
      if( keymap[keynote] ){
        // was on
        noteOff(keynote+C2, 0);
        keymap[keynote] = 0;
      }
    }
  }
}

/////////////////////////////////////////////////
int lastnote;

// monophonic note player.
//
void werk_noteout()
{
  if( data[curchain] != lastnote){
    if(data[curchain] == 0){
      noteOff(lastnote, 0);
      leds &= 0x7f;
    }else {
      noteOn(data[curchain], 127, 0);
      leds |= 0x80;
    }
    lastnote = data[curchain];
  }
  return;
}

// get the value from the werkstatt controls
// 0
// 1 lfo
// 2 freq
// 3 lfo wave
// 4 cutoff
// 5 pwm
// 
void werk_cc(int n)
{
   if( curchain == 0){
    return;
  }
  data[curchain] = ctrldata[n];
}

// get the note data.
// note this also reads the midi note data currently being played.
//
void werk_input_note(int s)
{ int n = notedata[0];


  if( curchain == 0){
    return;
  }
  if( s >= 0 && s < NCHAINS){
    if( notecnt == 0){
      startdata[s] = 0;
    }else {
      startdata[s] = n*2;
    }
    data[curchain] = startdata[s];
  }

}

// werkstatt gate
// 255 if notecnt > 0
// 0 if notecnt == 0
//
void werk_gate(int s)
{
  if( curchain == 0){
    return;
  }
  if( s >= 0 && s < NCHAINS){
    if( notecnt == 0){
      startdata[s] = 0;
    }else {
      startdata[s] = 255;
    }
    data[s] = startdata[s];
  }
}


// softbits interface to read midi data and play it.
// allows midi note data to be used the same as 
// keyboard note data
//
void werk_input_midi(int s)
{ int mcmd;

  if( curchain == 0){
    return;
  }

  mcmd = mstatus & 0xf0;
  if( mcmd == 0x90){  // note on
    if( mdata2 != 0  ){  // velocity 0 == note off
      noteOn(mdata1, mdata2, 0);
      
      if( midilastnote != mdata1 ){
        midilastnote = mdata1;
        data[curchain] = mdata1;  // note on
        
        if( s == 0){
          startdata[curchain] = data[curchain]; //
        }else {
          startdata[s] = data[curchain]; //
        }
        midigate = mdata2 << 1;
      }
      mstatus = 0;
    }else {
      mcmd = 0x80;    // velocity of 0 is note off
    }
  }
  if( mcmd == 0x80){
    noteOff(mdata1, 0);

    if( midilastnote == mdata1){
      data[curchain] = 0;  // note off
        if( s == 0){
          startdata[curchain] = 0; //
        }else {
          startdata[s] = 0; //
        }
      midigate = 0;
      midilastnote = 0;
    }
  }
  mstatus = 0;
}

#ifdef WDEBUG
void werk_out_midicc(int n)
{
  
}
#else

void werk_out_midicc(int n)
{
  out_midicc(n);  
}

#endif

////////////////////////////////////////////////////////////////////
// wrkstatt main processing function.
// drives the keyboad scan counter and 
// the hardware that uses it.
//

void werkstatt_sub()
{ int kin;
  int x;
  int ld;

  doclk();

  // curnote is even...
  if( curnote == count){
    if( curchan != 16){
      pinMode(MUXOUT, OUTPUT);
      digitalWrite(MUXOUT, HIGH);
      digitalWrite(MUXOUT, LOW);
      digitalWrite(MUXOUT, HIGH);
      digitalWrite(MUXOUT, LOW);
      if( muxout == 0){
        pinMode(MUXOUT, INPUT);
      }
    }
  }
   
}

void loop() 
{
  input_midi_sub(); // process any midi data..
  werkstatt_sub();  // run the werkstatt specific tasks

  // start softbits.
  power_on(1);

  midi_cc(19, 2);   // send CC 19 to chain 2.

  midi_chan1(0);

  werk_input_midi(0);    // get note on/off data
  // werk_noteout();   // output note data
  
  power_off();
  
  power_on(2);
  output_a5();      // send CC 19 (cutoff) data to pin 5 as PWM.
  power_off();

#ifdef ANALOGIN
// these need their own chains.
  power_on(3);
  werk_cc(4);     // cutoff
  werk_out_midicc(19);
  power_off();

  power_on(4);
  werk_cc(2);     // vco freq
  werk_out_midicc(10);
  power_off();

  power_on(5);
  werk_cc(1);     // lfo rate
  werk_out_midicc(3);
  power_off();
#endif  

  power_on(6);
//  werk_cc(3);
  werk_input_note(0);
  output_ax(6);

  power_off();
}

