// Softmidi
// Softmidi implements a number of Softbits that
// can be used to process Midi messages and data
//
// 11/16/2014 - initial release
// 1/5/2015 - first major release
// http://moddersandrockers.com/littlebits/softbitslive
//
// most midi msgs use 7 bit data. where it is
// 14 bit, only top 8 of 14 are used.
//
//

#include "softbits.h"
#include "midibits.h"
#include "soundbits.h"

int prevdata[NCHAINS];
unsigned char sysexbuf[128];
int wrcnt;
int wrptr;

int hexcnt = 0;

void debugmsg(char *str, int ln)
{
#ifdef DEBUG
#ifdef LEONARDO
  if( Serial){
#endif 
  if( ln){
    Serial.println(str);
    hexcnt = 0;
  }else {
    Serial.print(str);
  }
#ifdef LEONARDO
  }
#endif  
#endif
}

void debugInt(int n, int ln)
{
#ifdef DEBUG
#ifdef LEONARDO
  if( Serial){
#endif 
  if( ln){
    Serial.println(n);
    hexcnt = 0;
  }else {
    Serial.print(n);
    Serial.print(" ");
  }
#ifdef LEONARDO
  }
#endif  
#endif
}


#ifdef ETHERBIT

#include <SPI.h>
#include <Ethernet.h>

#include "etherbit.h"

// configure the Ethernet (mac , PORT ) in etherbit.h

EthernetServer server(PORT);
EthernetClient ether_client;
#endif

#ifdef YUNBIT

#include <Bridge.h>
#include <YunServer.h>
#include <YunClient.h>

// Initialize the Ethernet server library
// with the IP address and port you want to use
// (port 80 is default for HTTP):
YunServer server;
YunClient ether_client;
#endif
int ether_ok = 0;

void ether_close(char *msg)
{
#ifdef YUNBIT  
    ether_client.stop();
#else
#ifdef ETHERBIT
    ether_client.stop();
#endif
#endif
    debugmsg(msg, 1);
    ether_ok = 0;
}    

extern void xmessage( char *msg);
extern int maxchainnum;
extern void send_cur_data();

#define MIDISCALE 6

int mididevs;
int midiclock;


#define MBUFSIZ 8

struct midi_rec {
// input stuff  
int flags;      // used as a score board.  
int action; // current running status 
unsigned char mididata[3];
int midiidx;
int midicnt;
int inp, outp;
unsigned char buf[MBUFSIZ][3];
// output stuff
};


struct midi_rec midirxtx, midiusb;
struct midi_rec midi_ether;
struct midi_rec *midicurdev;


// softbit variables.
unsigned char midichan[NCHAINS];

//
void midi_setup(int n)
{
#ifdef LEONARDO  
  if( n == 0){
    Serial.begin(31250);  // RXTX
  }else if( n == 1){
    Serial1.begin(31250);  // USB
    mididevs |= 2;
  }
#else
  if( n == 0){
    Serial.begin(31250);
    mididevs |= 1;
  }else if( n == 1){
    Serial.begin(19200);  // normal serial
    mididevs |= 2;
  }
#endif

#ifdef ETHERBIT
  Ethernet.begin(mac, ip);
  server.begin();
  debugmsg("Start server", 1);
#endif

#ifdef YUNBIT
  Bridge.begin();
  
  server.listenOnLocalhost();
  server.begin();
  debugmsg("Start Yun server", 1);
#define ETHERBIT 2  
#endif

}

// output midi data.
// we want to trickle out the serial data so that the
// main loop timing is minimally impacted.
//
// midiwr, just write the data..
//
void midiwr(unsigned char *buf, int len)
{
#ifdef LEONARDO
    if( mididevs & 1){    
        Serial1.write(buf, len);
    }else if( (mididevs & 2) && Serial){  // USB
        Serial.write(buf, len);
     //   flash();
    }
#else  
    if( (mididevs & 1)){
      Serial.write(buf, len);
    }else if( mididevs & 2){
      Serial.write(buf, len);
    }
#endif  
}

// wrserial, trickle the bytes out...
//
void wrserial()
{ int cnt;
  int len;
  unsigned char *buf;
 
 buf = &sysexbuf[wrptr];
 len = 1;
 if( wrcnt - wrptr > 4){
   len = 4;
 }

  if( wrcnt > 0 ){
    midiwr( buf, len);
    wrptr += len;
    if( wrptr >= wrcnt){
      wrcnt = 0;
      wrptr = 0;
    }
  }
}

void midiwrite( unsigned char *buf, int len)
{  unsigned char *p, *q;

#ifndef ETHRBIT
  if( len + wrcnt < 128){
    p = &sysexbuf[wrcnt];
    q = buf;
    wrcnt += len;
    while(len > 0){
      *p++ = *q++;
      len--;    
    }
    return;
  }
#endif  
  if( midicurdev == &midi_ether){
    debugmsg("EMW ", 0); debugInt(len, 1);
    if( ether_ok ){
#ifdef ETHERBIT     
      ether_client.write(buf, len);
#endif      
    }
  }else {
    // to big to trickle so just write everything.
    if( wrcnt > 0){
      midiwr(&sysexbuf[wrptr], wrcnt-wrptr);
      wrcnt = 0;
      wrptr = 0;
    }
    midiwr( buf, len);
  }
  
}

// output a Control Change msg
// 7 bit value
//
void out_midicc(int cchan)
{ unsigned char buf[6];
  int val;

  if( curchain == 0){
    return;
  }
  val = data[curchain];
  if( prevdata[curchain] == val){
    return;
  }
  prevdata[curchain] = val;

  buf[0] = 0xb0 | (midichan[curchain] & 0xf);            // control change
  buf[1] = cchan & 0x7f;
  buf[2] = (val >> 1) & 0x7f;

  midiwrite(&buf[0], 3);
}

// output a Control Change msg
// 14 bit value, usually filter cutoff, modulation depth.
//
void output_midicc14( int cchan)
{ int val;
  unsigned char buf[6];
 
  if( curchain == 0){
    return;
  }
  val = data[curchain];
  val = data[curchain];
  if( prevdata[curchain] == val){
    return;
  }
  prevdata[curchain] = val;

  buf[0] = 0xb0 | (midichan[curchain] & 0xf);            // control change
  buf[4] = cchan & 0x7f;
  buf[5] = (val >> 1) & 0x7f;
  
  buf[3] = 0xb0 | (midichan[curchain] & 0xf);            // control change
  buf[1] = (cchan+32) & 0x7f;
  buf[2] = (val & 1) << 6;
  
  midiwrite(&buf[0], 6);
}

// output a pitch bend 
//

void output_midibend()
{ int val;
  unsigned char buf[6];

  if( curchain == 0){
    return;
  }
  val = data[curchain];

  buf[0] = 0xe0| (midichan[curchain] & 0xf); // pitchbend
  buf[1] = (val & 1) << 6;
  buf[2] = (val >> 1) & 0x7f;
      
  midiwrite(&buf[0], 3);
}

////////////////// MIDI POLL //////////////////
// read midi data
//

int val;
int midilastnote;    // for simple 
int midicurchan;    // current channel
int midigate;

void mouthex(int n)
{ char *tab= "0123456789ABCDEF";
#ifdef DEBUG
#ifdef LEONARDOX
  Serial1.print(tab[ 0xf & (n >> 4)] );
  Serial1.print(tab[ 0xf & n]);
  Serial1.print(" ");
#else
  Serial.print(tab[ 0xf & (n >> 4)] );
  Serial.print(tab[ 0xf & n]);
  Serial.print(" ");
#endif
  hexcnt++;
  if( hexcnt >= 16){
    Serial.print("\n");
    hexcnt = 0;
  }
#endif  
}

extern unsigned char prog0[];

int seql, seqh;
unsigned char touched[NCHAINS];
int msg_done = 0;    // used by the etherbit logic to know when to close the connection.
int recvticks;

////////////////////////////////////////////////////


void send_cur_data()
{  int i = 0;
  unsigned int d;
  unsigned char buf[8] = { 0xf0, 'S', 'B', 'L', 6, 0, 0, 2 };
  
#ifdef DEBUG
debugmsg("SC", 1);
#endif

  buf[5] = seqh;
  buf[6] = seql;
  
  if( maxchainnum < 2){
    maxchainnum = 2;
  }

  midiwrite(buf, 8);    
  for(i = 1; i <= maxchainnum; i++){
    d = 0xff & data[i];
    buf[0] = i  & 0x7f;
    buf[1] = (i >> 7 ) & 0x7f;
    buf[2] = d & 0x7f;
    buf[3] = (d >> 7 ) & 0x7f;
    midiwrite(buf, 4);
  }

  buf[0] = 0xf7;
  midiwrite(buf, 1);  

  if( midicurdev == &midi_ether){
    ether_close("SCD");
  }
  msg_done = 2;
}



void xmessage(char *msg)
{  int i = 0;
  unsigned char buf[8] = { 0xf0, 'S', 'B', 'L', 2, 0, 0, 0 };
  buf[5] = seqh;
  buf[6] = seql;

  midiwrite(buf, 8);

  while(*msg && i < 20){
    buf[0] = *msg++;
    midiwrite(buf, 1);
    i++;
  }
  buf[0] = 0xf7;
  midiwrite(buf, 1);  

  // Serial1.println(msg);
  if( midicurdev == &midi_ether){
    ether_close("SXM");
  }

}

////////////////////////////////////////////////////
void wire_recv()
{
  recvticks++;
  if( recvticks >= 100){
    recvticks = 0;
    send_cur_data();
  }
}

////////////////////////////////////////////////////

void processSysex(int len)
{ int ver;
  int i;
  unsigned char *p, *q;
  int p6, d6;


  if( sysexbuf[0] != 0x53 ||
    sysexbuf[1] != 0x42 ||
    sysexbuf[2] != 0x4c ){
      msg_done = 2;      // no reply possible..
#ifdef DEBUG
debugmsg("BH", 1);    // bad header
#endif
    return;
  }

  ver = sysexbuf[6];

  if( sysexbuf[3] == 1){
    seqh = sysexbuf[4];
    seql = sysexbuf[5];

    p = &sysexbuf[7];
    q = &prog0[0];
    for(i=7; i < len; i++){
	*q++ = *p++;
    }
    *q++ = 0xff;
    msg_done = 2;      // program load does not need reply
  }else if( sysexbuf[3] == 6){
    // reset prev startdata that we have touched.
    for(i=0; i < NCHAINS; i++){
      if( touched[i] != 0){
        startdata[i] = 255;
      }
      touched[i] = 0;
    }

    p = &sysexbuf[7];
    for(i=7; i < len; i+= 4){
      p6 = *p | (p[1] << 7);
      d6 = p[2] | (p[3] << 7);
      if( p6 > 0 && p6 < NCHAINS){
        touched[p6] = 1;
        startdata[p6] = d6;
        data[p6] = d6;
      }
      p = &p[4];
    }
    msg_done = 1;
  }
}

// process the byte read from the serial interface.

void process_midi( struct midi_rec *mr)
{ int inp;
  int outp;
  int mtmp;
  unsigned char *mdata;
  int i;

  if( val & 0x80){
    // control
    mtmp = val & 0xf0;
    if( mtmp == 0xf0){
      if ( val & 0x8){
        // f8 - ff realtime
        if( val == 0xf8){
          // timing clock
          midiclock++;
        }else if( val == 0xff){
          // reset
          mr->mididata[0] = val;    // clear running status
          mr->action = 0;
        }
      }else {
        // f0 - f7
        mr->mididata[0] = val;    // clear running status
        mr->midicnt = 0;

        if( val & 0x04){
          // no args f4-f7
          if( val == 0xf7){  // end sysex
#ifdef DEBUG
debugmsg("EX", 1);
#endif
		processSysex(mr->midiidx);
          }
          mr->action = 0;  // noop
          mr->midiidx = 0;
        }else {
          if( val == 0xf0){  // sysex
            mr->action = 1;
            mr->midiidx = 0;  // 0xf0 S B L X N N V X=cmd
            mr->midicnt = 128;
#ifdef DEBUG
debugmsg("SX", 1);
#endif
          }else {
          // else f1-f3
            mr->midiidx = 0;
            mr->action = 0;
          }
        }
      }
      return; // Fx
    }
    // basic state init for others.
    mr->mididata[0] = val;
    mr->action = mtmp;
    mr->midiidx = 1;
    mr->midicnt = 3;

    if(mtmp == 0xc0 || mtmp == 0xd0){  
      // program change and channel pressure only have 1 
      // data byte
      mr->midicnt = 2;
    }
    return;
  }else {
/////////////////////////// PROCESS DATA BYTES ////////////////////////////        
  // data
    if( mr->action == 0){
      return;
    }
    if( mr->action == 1){  // sysex data
	  sysexbuf[mr->midiidx] = val;
      if( mr->midiidx < 128){
        mr->midiidx++;
      }
      return;
    }
    mr->mididata[mr->midiidx] = val;
    mr->midiidx++;
    if( mr->midiidx < mr->midicnt){
      return;
    }
    
    msg_done = 2;    // flag that the ehterbit can disconnect

    // ok got enough data to process
    mr->midiidx = 1;  // assume running status
    // check the score board for this interface
    mtmp = ( mr->mididata[0] >> 4) & 7;
    mr->flags |= 3;    // note on and note off always
    if( (mr->flags & (1 << mtmp) ) == 0){
      return;  // do not need this message
    }
    
    inp = mr->inp;
    mdata = &mr->buf[inp][0];
    mdata[0] = mr->mididata[0];
    mdata[1] = mr->mididata[1];
    mdata[2] = mr->mididata[2];

    inp++;
    if( inp == MBUFSIZ){
      inp = 0;
    }
    mr->inp = inp;
  }  
}

// the main routines for midi processing.
//
// mididevs bottom 4 bits used to show which 
// midi inputs are being used
int mchan;
int mstatus;
int mdata1;
int mdata2;

#ifdef FLASH
int flashcnt;
int flashrate=2;

void flash()
{
  flashcnt++;
  if( flashcnt > flashrate/2){
    digitalWrite(5, HIGH);
  }
  if( flashcnt >= flashrate){
    digitalWrite(5, LOW);
    flashcnt = 0;
  }
  
}
#endif

int b64cnt;
int b64buf[4];
int val64;

int base64decode(int x)
{ int v;

  if( x >= 'A' && x <= 'Z'){
    v = x - 'A';
  }else if( x >= 'a' && x <= 'z'){
    v = x - 'a'+26;
  }else if( x >= '0' && x <= '9'){
    v = x - '0'+52;
  }else if( x == '+'){
    v = 62;
  }else if( x == '/'){
    v = 63;
  }else if( x == '='){
    b64cnt = 0;
    return 0;
  }else {
    b64cnt = 0;
    return 0;
  }

  if( b64cnt == 0){
    b64buf[0] = 0xfc & (v << 2);
    b64cnt++;
    return 0;
  }
  if( b64cnt == 1){
    b64buf[0] |= (v >> 4) & 0x03;
    b64buf[1] =  0xf0 & (v << 4);
    val64 = b64buf[0];
    b64cnt++;
    return 1;
  }
  if( b64cnt == 2){
    b64buf[1] |= 0x0f & ( v >> 2);
    b64buf[2] = 0xc0 & (v << 6);
    val64 = b64buf[1];
    b64cnt++;
    return 1;
  }
  if( b64cnt == 3){
    b64buf[2] |= v;
    b64cnt = 0;
    val64 = b64buf[2];
    return 1;
  }
  return 0;
}

// input_midi_sub
// this is not a softbit...
// input_midi is the softbit.

void input_midi_sub()
{ int nxt, cnt, i;
  struct midi_rec *mr;
  unsigned char *mdata;
  uint8_t buf[4];
  
  if( wrcnt > 0){
    wrserial();
    return;  // wait for output to be written.
  }
#ifdef LEONARDO
  while(1){
    if( (mididevs & 1) && Serial1.available() > 0) {
    // read the incoming byte:
      val = Serial1.read();
      process_midi( &midirxtx);
      if( msg_done == 1){
        send_cur_data();
      }
    }
    // check USB
    if( (mididevs & 2) && Serial && Serial.available() > 0) {
      // read the incoming byte:
      val = Serial.read();
      process_midi( &midiusb);
      if( msg_done == 1){
        send_cur_data();
      }
    }
    break;
  }
#else
  while(1){
    if( (mididevs & 3) && Serial.available() > 0) {
  // read the incoming byte:
      val = Serial.read();
      process_midi( &midirxtx);
      if( msg_done == 1){
        send_cur_data();
      }
    }
    break;
  }
#endif

#ifdef ETHERBIT
  if( ether_ok == 0){
#ifdef YUNBIT
  ether_client = server.accept();
#else 
  ether_client = server.available();
#endif  

    if( ether_client){
      ether_ok = 1;
      debugmsg("NC", 1);
    }
  }
  if (ether_client && ether_ok) {
    msg_done = 0;
    // read midi data from the connection.
    while((cnt = ether_client.available()) > 0 ){
      if( cnt > 4){
        cnt = 4;
      }
      val = ether_client.read(&buf[0], cnt);
      for(i=0; i < cnt; i++){
#ifdef USEBASE64        
        if( base64decode( buf[i]) ){
          val = val64;
          mouthex(val);
          process_midi( &midi_ether);
        }
#else 
          val = buf[i];
          mouthex(val);
          process_midi( &midi_ether);
#endif
        if( msg_done){
          i = -1;  // force break;
          break;
        }  
      }
      if( i != cnt){
        break;
      }
    }
    if( msg_done == 1){
      send_cur_data();
    }else if( msg_done == 2){
      ether_close("IMS");
    }
    mididevs |= 4;
  }
#endif
  // let chain process the data
  // do a round robin scheduler
  mstatus = 0;      // clear status == nothing to do.
  
  // find an active midi device
  for(nxt = 0; nxt < 3; nxt++){
    if( (mididevs & ( 1 << nxt) ) ){
      mr = &midirxtx;
#ifdef LEONARDO      
      if( nxt == 1){
        mr = &midiusb;
      }
#endif
#ifdef ETHERBIT
      if( nxt == 2 && (mididevs & 4) ){
        mr = &midi_ether;
      }
#endif
      midicurdev = mr;
      if( mr != NULL && mr->outp != mr->inp){
        // have data to process.
        mdata = &mr->buf[mr->outp][0];
        mstatus = mdata[0];
        mdata1 = mdata[1];
        mdata2 = mdata[2];
        mchan = (mdata[0] & 0x0f) ; // || ( 1 << mnext + 4);

        mr->outp++;
        if( mr->outp == MBUFSIZ){
          mr->outp = 0;
        }
        // flash();
        break;  // exit the loop
      }      
    }
  }
  
  // do simple midi processing
  if( mstatus == 0){
    return;
  }

  if( midicurchan && 
    (midicurchan - 1) != mchan){
      return; // simple channel mismatch
  }
}

// softbits interface
void input_midi(int s)
{ int mcmd;

  if( curchain == 0){
    return;
  }

  mcmd = mstatus & 0xf0;
  if( mcmd == 0x90){  // note on
    digitalWrite(13,LOW);
    if( mdata2 != 0  ){  // velocity 0 == note off
      if( midilastnote != mdata1 ){
        midilastnote = mdata1;
        data[curchain] = mdata1;  // note on
        
        if( s == 0){
          startdata[curchain] = data[curchain]; //
        }else {
          startdata[s] = data[curchain]; //
        }
        midigate = mdata2 << 1;
      }
    }else {
      mcmd = 0x80;    // velocity of 0 is note off
    }
  }
  if( mcmd == 0x80){
    digitalWrite(13,HIGH);
    if( midilastnote == mdata1){
      data[curchain] = 0;  // note off
        if( s == 0){
          startdata[curchain] = 0; //
        }else {
          startdata[s] = 0; //
        }
      midigate = 0;
      midilastnote = 0;
    }
  }
}

///
// select only channel 1
void midi_chan1(int s)
{
  if( curchain == 0){
    return;
  }
  if( mchan != 0){
    curchain = 0;
    if( s >= 0 && s < NCHAINS){
      startdata[s] = data[curchain];
    }
  }else {
    midicurchan = 1;
  }
}

//

void midi_notes(int s)
{
  if( curchain == 0){
    return;
  }
}

////
void midi_cc(int cont, int s)
{ int mcmd;

  if( curchain == 0){
    return;
  }
  midicurdev->flags |= 0xc;  // control change
  mcmd = mstatus & 0xf0;
  if( mcmd == 0xb0){

    if( mdata1 == cont){
      if( s >= 0 && s < NCHAINS){
        startdata[s] = mdata2 << 1;
        data[s] = startdata[s];
      }
    }
  }
}

void midi_cc19(int s)
{
  midi_cc(19, s);
}

// attack
//
void midi_attack(int s)
{
  midi_cc(73, s);
}

// release
//
void midi_release(int s)
{
  midi_cc(72, s);
}

// midi gate
// send whether a note is playing or not.
//
void midi_gate(int s)
{
  if( curchain == 0){
    return;
  }
  if( s >= 0 && s < NCHAINS){
    if( midigate == 0){
      startdata[s] = 0;
    }else {
      startdata[s] = 255;
    }
    data[s] = startdata[s];
  }
}

// 
void midi_velocity(int s)
{
  if( curchain == 0){
    return;
  }
  if( s >= 0 && s < NCHAINS){
    if( midigate != 0){
      startdata[s] = midigate;
    }
    data[s] = startdata[s];
  }
}

////
//
void midi_glide(int n)
{ int stp;
  unsigned int before;
  unsigned int cur;
    
  if( curchain == 0){
    return;
  }
  stp = 4;
  if( n > 0 && n < NCHAINS){
    stp = make8bit( data[n]);
  }
  
  cur = data[curchain];
  
  if( cur > 20){
    cur = cur << 8;
  
    before = counters[curchain];
    
    if( before < cur){
      before += stp;  // count up to cur.
      if( before > cur){
       before = cur;  // reached the note.
      }
    }else if( before > cur){
      before -= stp;
      if( before < cur){
       before = cur;
      }
    } 
    counters[curchain] = before;
    data[curchain] = (counters[curchain] >> 8) & 0xff;
  }else {
    data[curchain] = 0;    // squelch below 20
  }
}


//////////////////////////////////////
/// raw serial io

void input_serial()
{  int val = 0;

  if( curchain ==0){
    return;
  }
  
#ifdef LEONARDO
  while(1){
    if( (mididevs & 1) && Serial1.available() > 0) {
    // read the incoming byte:
      val = Serial1.read();
    }else if( (mididevs & 2) && Serial && Serial.available() > 0) {
      // read the incoming byte:
      val = Serial.read();
    }else {
      curchain = 0;
    }
    break;
  }
#else
  while(1){
    if( (mididevs & 3) && Serial.available() > 0) {
  // read the incoming byte:
      val = Serial.read();
    }else {
      curchain = 0;
    }
    break;
  }
#endif

  data[curchain] = val;
}

void output_serial()
{  unsigned char buf[2];

  if( curchain == 0){
    return;
  }
  
  buf[0] = data[curchain];
  midiwrite(buf, 1);
}

////////////////////////////////////////

void midi_out_noteon()
{
  
}


void midi_out_noteoff()
{
  
}

////////////////////////////////////////

