// 8 bit softbits
// this program implements some simple modules that can
// be used as virtual bits.
//
// 11/16/2014 - initial release
// http://moddersandrockers.com/littlebits/2.php
//
// The LittleBits snap together modules use a simple
// interface for all input and output so that if the
// modules snap together then you will not break anything.
//
// These soft bits also use a simple interface that allows
// them to be combined in a number of ways and it will not
// break the program.
//
// the softbits form chains of modules.
// The first module of a chain is a power module
// which also specifies the chain number.
// chain 0 is the do nothing chain.

#include "softbits.h"

// define the data storage used.
int data[NCHAINS];    // allow for 20 chains.
int chain;      // used to hold the active chain number.
int curchain;   // the current chain value
int startdata[NCHAINS]; // power uses this value as start value.
int counters[NCHAINS]; // counters - 16 bit, top 8 bits is output
int lastout[20];
int pinmap;
unsigned int latchclk[2];    // allow for 32 latches, each bit is prev clk for chan 
unsigned int latchdata[2];    // allow for 32 latches, each bit is prev clk for chan 

// define some utility functions
int make8bit(int n)
{
  if( n > 255){
    return 255;
  }
  if( n < 0){
    return 0;
  }
  return n;
}

void input_a0()
{
  if( curchain==0){
    return;
  }
  // read and convert down to 8 bits.
  data[curchain] = analogRead(0) >> 2;
}

void input_a1()
{
  if( curchain==0){
    return;
  }
  data[curchain] = analogRead(1) >> 2;
}

void input_ax(int n)
{
  if( curchain==0){
    return;
  }
  
  data[curchain] = analogRead(n) >> 2;
}

void input_d0()
{
  input_dx(0);
}

void input_dx(int n)
{
  if( curchain==0){
    return;
  }
  if( pinmap & (1<< n)){
    pinMode(n, INPUT);
    pinmap &= ~(1 << n);
  }
  data[curchain] = digitalRead(n) ? 255 : 0;
 }

// analog outputs assume that the switch is
// set to analog
void output_a5()
{
  if( curchain==0){
    return;
  }
  if( lastout[5] != data[curchain]){
    analogWrite(5, data[curchain]);
    lastout[5] = data[curchain];
  }
}  

void output_a9()
{
  if( curchain==0){
    return;
  }
  if( lastout[9] != data[curchain]){
    analogWrite(9, data[curchain]);
    lastout[9] = data[curchain];
  }
}  

void output_a11()
{
  if( curchain==0){
    return;
  }
  if( lastout[11] != data[curchain]){
    analogWrite(11, data[curchain]);
    lastout[11] = data[curchain];
  }
}  

void output_ax(int n)
{
  if( curchain==0){
    return;
  }
  if( n >= 0 && n < 20){
    if( lastout[n] != data[curchain]){
      analogWrite(n, data[curchain]);
      lastout[n] = data[curchain];
    }
  }
}  

// output digital data. Assumes that the switch is 
// in the pwm position.
void output_d5()
{
  output_dx(5);
}

void output_d9()
{
  output_dx(9);
}
/////

void output_d10()
{
  output_dx(10);
}

void output_d11()
{
  output_dx(11);
}


void output_d12()
{
  output_dx(12);
}

void output_d13()
{
  output_dx(13);
}


void output_d1()
{
  output_dx(1);
}

void output_dx(int n)
{
  if( curchain==0){
    return;
  }
  if( (pinmap & ( 1 << n)) == 0){
    pinMode(n, OUTPUT);
    pinmap |= (1 << n);
  }
  if( data[curchain] >= 128){
    digitalWrite(n, HIGH);
  }else {
    digitalWrite(n, LOW);
  }
}

// special oscillator output bits to stop the chirp.
void oscillator_squelch()
{  int tmp;
  if( curchain==0){
    return;
  }
  tmp=make8bit(data[curchain]);
  if( tmp < 10){
    tmp = 0;
  }else if( tmp < 30){
    tmp = 30;
  }
  data[curchain] = tmp;
}  

void midi_cv(int root)
{  int tmp;
  if( curchain==0){
    return;
  }
  if( root > 0){
  }
  tmp=make8bit(data[curchain]);
  if( root > 0){
    tmp -= root-12;
  }
  if( tmp < 0){
    tmp = 0;
  }
  
  tmp = tmp + tmp;
  tmp= tmp + tmp + tmp;
  if( tmp < 0){
    tmp = 0;
  }else if( tmp > 255){
    tmp = 255;
  }
  data[curchain] = tmp;
}  


// logic bits
//
void logic_and(int n)
{ int data2 = 255;
  if( curchain==0){
    return;
  }
  if( n > 0 && n < NCHAINS){
    data2 = data[n];
  }
  
  if( data2 >= 128 && data[curchain] >= 128){
    data[curchain] = 255;
  }else {
    data[curchain] = 0;
  }
}

void logic_nand(int n)
{ int data2 = 255;
  if( curchain==0){
    return;
  }
  if( n > 0 && n < NCHAINS){
    data2 = data[n];
  }
  
  if( data2 >= 128 && data[curchain] >= 128){
    data[curchain] = 0;
  }else {
    data[curchain] = 255;
  }
}


void logic_or(int n)
{ int data2 = 0;
  if( curchain==0){
    return;
  }
  if( n > 0 && n < NCHAINS){
    data2 = data[n];
  }
  
  if( data2 >= 128 || data[curchain] >= 128){
    data[curchain] = 255;
  }else {
    data[curchain] = 0;
  }
}


void logic_nor(int n)
{ int data2 = 0;
  if( curchain==0){
    return;
  }
  if( n > 0 && n < NCHAINS){
    data2 = data[n];
  }
  
  if( data2 >= 128 || data[curchain] >= 128){
    data[curchain] = 0;
  }else {
    data[curchain] = 255;
  }
}

void logic_xor(int n)
{ int dat = data[curchain];
  int data2 = 255 - dat;

  if( curchain==0){
    return;
  }

  if( n > 0 && n < NCHAINS){
    data2 = data[n];
  }
  
  if( dat > 127 && data2 > 127){  // both high
    data[curchain] = 0;
  }else if( dat < 128 && data2 < 128){  // both low
    data[curchain] = 0;
  }else {
    data[curchain] = 255;
  }  
}

// logic_not
// input >= 128 output 0
// else output 255
//
// 
void logic_not()
{
  if( curchain==0){
    return;
  }
  if( data[curchain] >= 128 )
  {
    data[curchain] = 0;
  }else {
    data[curchain] = 255;
  }
}

// logic_compare
//
void logic_compare(int n)
{ int tmp;

  if( curchain==0 ){
    return;
  }
  tmp = data[curchain];
  if(n > 0 &&  n < NCHAINS){
    tmp = make8bit(data[n]);
  }
  
  if( data[curchain] > tmp){
    data[curchain] = 255;
  }else {
    data[curchain] = 0;
  }
}

// logic_compare
//
void logic_latch(int n)
{ int tmp;
  int clk;
  unsigned int prevclk;
  unsigned int mask;
  int idx;

  if( curchain==0 ){
    return;
  }
  clk = data[curchain];
    
  if( n > 0 && n < NCHAINS){
    clk = make8bit(data[n]);
  }

  mask = 1 << (curchain & 0x0f);
  idx = (curchain >> 4);
  prevclk = latchclk[ idx ] & mask;
  tmp = latchdata[ idx] & mask;

  if( clk > 127 ){
    clk = mask;
  }else {
    clk = 0;
  }
  
  if( clk && prevclk == 0){
    // positive edge on clk;
    if( n > 0){
      // sample input
      if( data[ curchain] > 128){
        tmp = mask;
      }else {
        tmp = 0;
      }
    }else {
      // toggle
      if( tmp ){
        tmp = 0;
      }else {
        tmp = mask;
      }
    }
  }

  // update clk state
  latchclk[ idx] =  clk | ( latchclk[ idx] & ~mask);
  // update prev data
  latchdata[ idx]=  tmp | (latchdata[ idx] & ~mask);
  
  if( tmp){
    data[curchain] = 255;
  }else {
    data[curchain] = 0;
  }
}

// logic_if();
//   softbit();  // run if true
//   softbit();  // run if true
// logic_else();
//   softbit();  // run if false
// logic_endif();// end of an if or an else
// uses curchain to enable/disable following softbits
//
void logic_if()
{
  if( curchain==0){
    return;
  }
  if( data[curchain] < 128){
    curchain = 0;  // setting chain to 0 causes the
    // following softbits to do nothing
  }
}

void logic_else()
{
  if( curchain == 0){
      curchain = chain;
  }else {
    curchain = 0;
  }
}

void logic_endif()
{
  curchain = chain;
}

// wire_loop makes current value the start value
// next time. Same as
// wire_split(n) where n is the current chain number.

void wire_loop()
{
  if( curchain==0){
    return;
  }
  startdata[curchain] = data[curchain];
}

// logic_delay( n )
// use the counter to implement a delay 
//
// if input is high then delay is active
// chain specified is used as amount to delay
// amount == 255 means delay forever
//
// if input is low then reset the delay
//
void logic_delay(int n)
{  int ticks;

  if( curchain==0 ){
    return;
  }
  // check the value from the specified chain
  if( n >= 1 && n < NCHAINS){
    ticks = make8bit(data[n]);
  }else {
    ticks = 254;
  }
  
  ticks = ticks * 16;    // slow it down.
  
  if( data[curchain] >= 128){
    // delay is active
    if( ticks < (255 * 16) && counters[curchain] < ticks){
      counters[curchain]++;
      if( counters[curchain] >= ticks){
        // delay end
        counters[curchain] = ticks;
      }
    }
    if( counters[curchain] < ticks){
      // turn off rest of chain
      chain = 0;
      curchain = 0;
    }
  }else {
    // reset delay
    counters[curchain] = 0;
  }
}

//////////////////////
//
void arith_plus(int n)
{  int tmp;

  if( curchain==0 || n == 0 || n >= NCHAINS){
    return;
  }
  
  tmp = make8bit(data[n]);    // get the data from the other chain

  tmp = make8bit(data[curchain] + tmp);

  data[curchain] = tmp;
}


void arith_minus(int n)
{  int tmp;

  if( curchain==0 || n == 0 || n >= NCHAINS){
    return;
  }
  
  tmp = make8bit( data[n]);    // get the data from the other chain

  data[curchain] = make8bit( data[curchain] - tmp);
}


// invert the current value
// out of range input gets converted to valid output
//

void arith_invert()
{
  if( curchain==0){
    return;
  }
  data[curchain] = 255 - make8bit( data[curchain]);
}

void arith_inc()
{
  if( curchain==0){
    return;
  }
  data[curchain]++;
  if( data[curchain] >= 256){
    data[curchain] = 0;  // roll over
  }
}


void arith_dec()
{
  if( curchain==0){
    return;
  }
  data[curchain]--;
  if( data[curchain] < 0){
    data[curchain] = 255;  // roll over
  }
}

void arith_times(int n)
{  int tmp;

  if( curchain==0 || n == 0 || n >= NCHAINS){
    return;
  }
  
  tmp = make8bit( data[n]);    // get the data from the other chain
  tmp = data[curchain] * tmp;

  data[curchain] = make8bit(tmp);
}

// divide current value by specified chain value
//
// if dividing by 0 then return max (255)
//
void arith_divide(int n)
{  int tmp;
  int res;

  if( curchain==0 || n == 0 || n >= NCHAINS){
    return;
  }
  
  tmp = make8bit( data[n]);    // get the data from the other chain
  res = make8bit(data[curchain]) << 7;
  if( tmp > 0){
    tmp = res / tmp;
    data[curchain] = (tmp >> 7) & 0xff;
  }else {
    data[curchain] = 255;    // div by 0 so set max.
  }
}

void arith_diff(int n)
{  int tmp;
  int res;

  if( curchain==0 || n == 0 || n >= NCHAINS){
    return;
  }
  
  tmp = make8bit( data[n]);    // get the data from the other chain
  res = make8bit(data[curchain]);
  
  if( tmp > res){
    data[ curchain] = tmp - res;
  }else{
    data[ curchain] = res - tmp;
  }
}

// set the chain value to n
//
void arith_set(int n)
{
  if( curchain==0){
    return;
  }
  data[curchain] = make8bit(n);
}


// counter adds the input value to the counter 
// for this chain. Treat input of 128 as 0.
// treat 255 as +127 and 0 as - 128
// this gives a natural center off feel.
//
// if trig is non zero then when the counter passes 0
// set trig chain startdata to 255
//
void arith_counter(int trig)
{ int n;
  int before;

  if( curchain==0){
    return;
  }
  
  n = make8bit( data[curchain])-128;
  
  before = counters[curchain];
  counters[curchain] = before + n;
  
  // just use the top 8 bits as the value.
  data[ curchain] = (counters[curchain] >> 8) & 0xff;
  
  // if a trigger chain is set the look for going past 0.
  if( trig > 0 && trig < NCHAINS){
    if( n > 0 && before < 0 && counters[curchain] >= 0){
      // passed 0 counting up
      startdata[trig] = 255;
    }else if( n < 0 && before >= 0 && counters[curchain] < 0){
      // passed 0 counting down
      startdata[trig] = 255;
    }
  }
}


// start a chain and set the value to the current start value
//
int maxchainnum = 0;    // what was the max chain number.

void power_on(int n)
{
  if( n > 0 && n < NCHAINS){
    chain = n;
    curchain = n;
    data[n] = startdata[n];  // use start data.
    if( n > maxchainnum ){
      maxchainnum = n;
    }
  }
}

// following bits will be 'powered off'
// 
void power_off()
{
  chain = 0;
  curchain = 0;
}

// take current value and pass to a new chain
//
void wire_split(int n)
{
  if( curchain==0){
    return;
  }
  if( n > 0 && n < NCHAINS){
    startdata[n] = data[curchain];
  }
}

//////////////////////////////
// 
int debug_cnt;

void debug_trace(const char *msg)
{
  if( curchain==0){
    return;
  }
  debug_cnt++;
#ifdef SDEBUG  
  Serial.print(debug_cnt);
  Serial.print(" ");
  Serial.print(curchain);
  Serial.print(" ");
  Serial.print(msg);
  Serial.print(" ");
  Serial.println(data[curchain]);
#endif    
}


// some debug printing functions.
//
char hextab[] = "0123456789ABCDEF";

#ifdef SDEBUG
void outHex(int n){
int val = n >> 12;
Serial.print(hextab[val & 0x0f]);
val = n >> 8;
Serial.print(hextab[val & 0x0f]);
val = n >> 4;
Serial.print(hextab[val & 0x0f]);
Serial.print(hextab[n & 0x0f]);
Serial.print(" ");
}

void outInt( int n)
{
  Serial.print(n);
  Serial.print(" ");
}
#endif

// some softbits need to be setup before they can be used.
//

void use_d5()
{
  pinMode(5, OUTPUT);
}

void use_d9()
{
  pinMode(9, OUTPUT);
}

void use_d1()
{
  pinMode(1, OUTPUT);
}

void use_d0()
{
  pinMode(0, INPUT);
}

void use_d10()
{
  pinMode(10, OUTPUT);
}

void use_d11()
{
  pinMode(11, OUTPUT);
}

void use_d12()
{
  pinMode(12, OUTPUT);
}

void use_d13()
{
  pinMode(13, OUTPUT);
}

/////
//
void use_trace()
{
  Serial.begin(9600);
}

// define which bits you are going to use.
//
// this function is only called once after a reset.
void soft_setup() {
  int i;

// init the startdata to the max value.  
  for(i=0; i < NCHAINS; i++){
    startdata[i] = 255;
  }

}

